/* Copyright 2004 NetApp, Inc., contribution by Eric Gumtow.
 * Copyright 2008 NetApp, Inc., contribution by Ganesh R Narayanan.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License v2 for more
 * details.
 */ 

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/types.h>
#include <asm/uaccess.h>


#include "procfsfile_pvt.h"

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#ifdef MODULE
MODULE_AUTHOR("Eric Gumtow");
#endif

static struct proc_dir_entry *root;

static int reader_func(char *, char **, off_t, int, int *, void *);
static int writer_func(struct file *, const char *, unsigned long , void *);
void *rlm_proc_stat_buf_loc(char *);


static u_int8_t pktd_reader_buf[728];
static u_int8_t pktd_writer_buf[688];
static u_int8_t conlog_buf[80]; 
static u_int8_t downbeat_buf[48]; 
static u_int8_t upbeat_buf[16]; 
static u_int8_t ecc_buf[320];
static u_int8_t watcher_buf[520];
static u_int8_t trapd_buf[712];
static u_int8_t coll_diags_buf[156]; 
static u_int8_t ned_buf[192];
static u_int8_t rsad_buf[544];
static u_int8_t healthd_buf[360]; 


procfs_props_t files[] = {

	{"/proc/rlm/pktd_reader", pktd_reader_buf, 728},
	{"/proc/rlm/pktd_writer", pktd_writer_buf, 688},
	{"/proc/rlm/conlog", conlog_buf, 80},
	{"/proc/rlm/f2r", downbeat_buf, 48},
	{"/proc/rlm/r2f", upbeat_buf, 16},
	{"/proc/rlm/ecc", ecc_buf, 320},
	{"/proc/rlm/watcher", watcher_buf, 520},
	{"/proc/rlm/trapd", trapd_buf, 712},
	{"/proc/rlm/coll_diags", coll_diags_buf, 156},
	{"/proc/rlm/ne", ned_buf, 192},
	{"/proc/rlm/rsad", rsad_buf, 544},
	{"/proc/rlm/rs_healthd", healthd_buf, 360},
};


/* create each file in the procfs */
static int __init
procfs_init_files(void)
{
	int i;
	char *file_name;

	/* create the root directory under /proc */
	root = proc_mkdir(ROOT_NAME, NULL);
	root->owner = THIS_MODULE;

	for (i = 0; i < NUM_PROCFS_FILES; i++) {

		/* grab just the file name from the complete path */
		file_name = strstr(files[i].file_name, ROOT_NAME);
		if (file_name) {
			file_name += strlen(ROOT_NAME);
			file_name = strchr(file_name, '/');
			if (file_name) {
				file_name++;
			}
		}

		if (file_name && file_name[0] && 
		    files[i].buf && files[i].buf_size > 0) {

			/* create the file */
			files[i].entry = create_proc_entry(file_name, 0666, root);
			if (files[i].entry == NULL) {
				return -ENOMEM;
			}

			/* data can be used for anything.  we use it to find the 
			 * buffer and the buffer's size */
			files[i].entry->data = &(files[i]);
			files[i].entry->read_proc = reader_func;
			files[i].entry->write_proc = writer_func;
			files[i].entry->owner = THIS_MODULE;
		}
	}

	printk(KERN_INFO "%s %s initialized\n", MODULE_NAME, MODULE_VERSION);

	return 0;
}

static void __exit
procfs_remove_files(void)
{
	int i;

	for (i = 0; i < NUM_PROCFS_FILES; i++) {
		remove_proc_entry(files[i].file_name, root);
	}
	remove_proc_entry(ROOT_NAME, NULL);

	printk(KERN_INFO "%s %s removed\n", MODULE_NAME, MODULE_VERSION);
}

/* the same reader func for everybody.  we know what buffer to read from by
 * using the data parameter */
static int
reader_func(char *page, char **start, off_t off, int count, 
            int *eof, void *data)
{
	procfs_props_t *pfile = (procfs_props_t *) data;

	MOD_INC_USE_COUNT;

	/* read requested wrong number of bytes.  maybe the app was recompiled
	 * and this module was not and this module has the wrong size for the
	 * buffer */
	if (count != pfile->buf_size) {
		MOD_DEC_USE_COUNT;
		return -EINVAL;
	}

	if (!memcpy(page, pfile->buf, count)) {
		MOD_DEC_USE_COUNT;
		return -EINVAL;
	}

	MOD_DEC_USE_COUNT;

	return count;
}

/* the same writer func for everybody.  we know what buffer to read from by
 * using the data parameter */
static int
writer_func(struct file *file, const char *buffer, 
            unsigned long count, void *data)
{
	procfs_props_t *pfile = (procfs_props_t *) data;

	MOD_INC_USE_COUNT;

	/* write requested wrong number of bytes.  maybe the app was recompiled
	 * and this module was not and this module has the wrong size for the
	 * buffer */
	if (count != pfile->buf_size) {
		MOD_DEC_USE_COUNT;
		return -EINVAL;
	}

	if (copy_from_user(pfile->buf, buffer, count)) {
		MOD_DEC_USE_COUNT;
		return -EFAULT;
	}

	MOD_DEC_USE_COUNT;

	return count;
}

/* this function can be imported by other modules, if those modules
 * want to keep their stats their own way. */
void *
rlm_proc_stat_buf_loc(char *file_name)
{
	int i;

	for (i = 0; i < NUM_PROCFS_FILES; i++) {
		if (!strcmp(file_name, files[i].file_name)) {
			return files[i].buf;
		}
	}

	return NULL;
}


module_init(procfs_init_files);
module_exit(procfs_remove_files);

MODULE_DESCRIPTION("RLM procfs files");
