/* Copyright 2004 NetApp, Inc., contribution by Dan Gilson.
 * Copyright 2008 NetApp, Inc., contribution by Ganesh R Narayanan.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License v2 for more
 * details.
 */ 

#if defined(CONFIG_MODVERSIONS)
#define	MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/fs.h>

#ifdef	MODULE
#include <linux/module.h>	/* Kernel module definitions */
#endif
#include <linux/init.h>
#include <linux/kernel.h>	/* We will be "in the kernel" mode of execution */

#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/ibm4xx.h>
#include <linux/spinlock.h>

#include <linux/ibm_ocp_gpio.h>

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

void led_on(void);
void led_off(void);

static int  open_count    = 0;
static int  release_count = 0;
static int  read_count    = 0;
static int  write_count   = 0;
static int  ioctl_count   = 0;
static int  timer_count   = 0;
static long most_jiffies  = 0;
static long old_jiffies   = 0;

static struct timer_list  timer;
static char fSet = 0; /* this is either 0 or 1.  0 => ibm_gpio_clr and 1=> ibm_gpio_set */

static void timer_callback(unsigned long data)
{
  int erc;

  timer_count++;

  if (most_jiffies < (jiffies - old_jiffies) ) 
	  most_jiffies = jiffies - old_jiffies;

  fSet ^= 1;
  if (fSet) {
	  /* if ( (erc = ibm_gpio_out(0, 1 << (8 * sizeof(u32) - (LUNA_GPIO_WATCHDOG+1)), ~0)) ) */
	  if ( (erc = ibm_gpio_kern_set(LUNA_GPIO_WATCHDOG)) )
		  printk("ibm_gpio_kern_set returned erc=%d\n",erc);
  } else {
	  /* if ( (erc = ibm_gpio_out(0, 1 << (8 * sizeof(u32) - (LUNA_GPIO_WATCHDOG+1)), 0)) ) */
	  if ( (erc = ibm_gpio_kern_clr(LUNA_GPIO_WATCHDOG)) )
		  printk("ibm_gpio_kern_clr returned erc=%d\n",erc);
  }
  old_jiffies = jiffies;
  timer.expires = jiffies + (HZ / 2);
  add_timer(&timer);
   
}


static int luna_wdog_open(struct inode *inode, struct file *file)
{
  printk("luna_wdog_open called\n");
  open_count++;
  return 0;
}

static int luna_wdog_release(struct inode *inode, struct file *file)
{
  printk("luna_wdog_release called\n");
  release_count++;
  return 0;
}


static ssize_t luna_wdog_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
  printk("luna_wdog_read called\n");
  read_count++;
  return 0;
}

static ssize_t luna_wdog_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
  printk("luna_wdog_write called\n");
  write_count++;
  return 0;
}

static int luna_wdog_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
  printk("luna_wdog_ioctl called\n");
  ioctl_count++;
  return 0;
}


static struct proc_dir_entry *luna_wdog_dir, *parent_dir = NULL;

static int luna_wdog_read_proc(char *buf, char **start, off_t pos, int count, int *eof, void *data)
{
  char    *p = buf;

  p += sprintf(p, "Luna WatchDog Status: \n");
  p += sprintf(p, "\t calls to open: %d\n", open_count);
  p += sprintf(p, "\t calls to release: %d\n", release_count);
  p += sprintf(p, "\t calls to read: %d\n", read_count);
  p += sprintf(p, "\t calls to write: %d\n", write_count);
  p += sprintf(p, "\t calls to ioctl: %d\n", ioctl_count);
  p += sprintf(p, "\t number of timers called: %d\n", timer_count);
  p += sprintf(p, "\t current system jiffies: %lu\n", jiffies);
  p += sprintf(p, "\t most jiffies between calls: %lu\n", most_jiffies);

  return (p-buf);
}


static struct file_operations luna_wdog_fops = {
  read:           luna_wdog_read,
  write:          luna_wdog_write,
  ioctl:          luna_wdog_ioctl, 
  open:           luna_wdog_open,
  release:        luna_wdog_release, 
};



MODULE_AUTHOR("Dan Gilson");

int testparam = 0;
MODULE_PARM(testparam, "i"); /* Can be set through insmod and modprobe */
MODULE_PARM_DESC(testparam, "Test parameter for observing module behavior, 0 = default");


static int  major;
#define NAME  "lunawdog"

/* When the module is loaded, execute the driver initialization routine. */
static int __init lunawdog_init(void)
{
  printk("In lunawdog_init, testparam = %d.\n", testparam);

  /* Use major number of 0 to get one dynamically assigned. */
  major = register_chrdev(0, NAME, &luna_wdog_fops);
  printk("registered char device lunawdog with major number %d.\n", major);

  /* Install /proc interface */
  if ((luna_wdog_dir = create_proc_entry("lunawdog", 0, parent_dir)) == NULL) {
    printk("lunawdog : Unable to create /proc entry\n");
    return -ENOMEM;
  }
  else {
    luna_wdog_dir->read_proc = luna_wdog_read_proc;
  }

  /* Initialize timer */
  init_timer(&timer);
  timer.function = timer_callback;
  timer.data     = 0;

  old_jiffies = jiffies;
  /* Start timer to simulate interrupt input source */
  timer.expires = jiffies + (HZ / 2);
  add_timer(&timer);

  /* Initialize spinlock. */
  spin_lock_init(&lock);

  return(0);
}

/* Just before the module is removed, execute the stop routine. */
static void __exit lunawdog_stop(void)
{
  printk("In lunawdog_stop.\n");

  unregister_chrdev(major, NAME);

  remove_proc_entry("lunawdog", 0);

  /* Stop the timer */
  del_timer(&timer);

  return;
}

module_init(lunawdog_init);
module_exit(lunawdog_stop);
